# Copyright 2016-2024 Euratom
# Copyright 2016-2024 United Kingdom Atomic Energy Authority
# Copyright 2016-2024 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import json
import numpy as np
from cherab.core.utility import RecursiveDict
from cherab.core.atomic import Element
from .utility import DEFAULT_REPOSITORY_PATH, valid_charge, encode_transition

"""
Utilities for managing the local rate repository - PEC section.
"""


def add_pec_excitation_rate(element, charge, transition, rate, repository_path=None):
    """
    Adds a single PEC excitation rate to the repository.

    If adding multiple rate, consider using the update_pec_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param transition: Tuple containing (initial level, final level).
    :param rate: Excitation PEC dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with excitation PEC in photon.m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    update_pec_rates({
        'excitation': {
            element: {
                charge: {
                    transition: rate
                }
            }
        }
    }, repository_path)


def add_pec_recombination_rate(element, charge, transition, rate, repository_path=None):
    """
    Adds a single PEC recombination rate to the repository.

    If adding multiple rate, consider using the update_pec_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param transition: Tuple containing (initial level, final level).
    :param rate: Recombination PEC dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'rate': array-like of size (N, M) with recombination PEC in photon.m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    update_pec_rates({
        'recombination': {
            element: {
                charge: {
                    transition: rate
                }
            }
        }
    }, repository_path)


def add_pec_thermal_cx_rate(donor_element, donor_charge, receiver_element, receiver_charge, transition, rate, repository_path=None):
    """
    Adds a single PEC thermal charge exchange rate to the repository.

    If adding multiple rate, consider using the update_pec_thermal_rates() function
    instead. The update function avoid repeatedly opening and closing the rate
    files.

    :param donor_element: Electron donor plasma species (Element/Isotope).
    :param donor_charge: Electron donor charge.
    :param receiver_element: Electron receiver plasma species (Element/Isotope).
    :param receiver_charge: Electron receiver charge.
    :param transition: Tuple containing (initial level, final level).
    :param rate: Thermal CX PEC dictionary containing the following entries:

    |      'ne': array-like of size (N) with electron density in m^-3,
    |      'te': array-like of size (M) with electron temperature in eV,
    |      'td': array-like of size (K) with donor temperature in eV,
    |      'rate': array-like of size (N, M, K) with thermal CX PEC in photon.m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """
    rates2update = RecursiveDict()
    rates2update[donor_element][donor_charge][receiver_element][receiver_charge][transition] = rate

    update_pec_thermal_cx_rates(rates2update.freeze(), repository_path)


def update_pec_rates(rates, repository_path=None):
    """
    Updates excitation and recombination PEC files /pec/<class>/<element>/<charge>.json.
    in the atomic data repository.

    File contains multiple PECs, indexed by the transition.

    :param rates: Dictionary in the form:

    |  { <class>: { <element>: { <charge>: { <transition>: <pec> } } } }, where
    |      <class> is the one of the following PEC types: 'excitation', 'recombination'.
    |      <element> is the plasma species (Element/Isotope).
    |      <charge> is the charge of the plasma species.
    |      <transition> is the tuple containing (initial level, final level).
    |      <pec> is the PEC dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'rate': array-like of size (N, M) with PEC in photon.m^3.s^-1.

    :param repository_path: Path to the atomic data repository.
    """

    valid_classes = [
        'excitation',
        'recombination'
    ]

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for cls, elements in rates.items():
        for element, charge_states in elements.items():
            for charge, transitions in charge_states.items():

                # sanitise and validate
                cls = cls.lower()
                if cls not in valid_classes:
                    raise ValueError('Unrecognised pec rate class \'{}\'.'.format(cls))

                if not isinstance(element, Element):
                    raise TypeError('The element must be an Element object.')

                if not valid_charge(element, charge):
                    raise ValueError('Charge state is larger than the number of protons in the element.')

                path = os.path.join(repository_path, 'pec/{}/{}/{}.json'.format(cls, element.symbol.lower(), charge))

                # read in any existing rates
                try:
                    with open(path, 'r') as f:
                        content = RecursiveDict.from_dict(json.load(f))
                except FileNotFoundError:
                    content = RecursiveDict()

                # add/replace data for a transition
                for transition in transitions:
                    key = encode_transition(transition)
                    data = rates[cls][element][charge][transition]

                    # sanitise/validate data
                    data['ne'] = np.array(data['ne'], np.float64)
                    data['te'] = np.array(data['te'], np.float64)
                    data['rate'] = np.array(data['rate'], np.float64)

                    if data['ne'].ndim != 1:
                        raise ValueError('Density array must be a 1D array.')

                    if data['te'].ndim != 1:
                        raise ValueError('Temperature array must be a 1D array.')

                    if (data['ne'].shape[0], data['te'].shape[0]) != data['rate'].shape:
                        raise ValueError('Density, temperature and rate data arrays have inconsistent sizes.')

                    content[key] = {
                        'ne': data['ne'].tolist(),
                        'te': data['te'].tolist(),
                        'rate': data['rate'].tolist()
                    }

                # create directory structure if missing
                directory = os.path.dirname(path)
                if not os.path.isdir(directory):
                    os.makedirs(directory)

                # write new data
                with open(path, 'w') as f:
                    json.dump(content, f, indent=2, sort_keys=True)


def update_pec_thermal_cx_rates(rates, repository_path=None):
    """
    Updates thermal CX PEC files /pec/thermal_cx/<donor_element>/<donor_charge>/<receiver_element>/<receiver_charge>.json
    in the atomic data repository.
    
    File contains multiple PECs, indexed by the transition.
    
    :param rates: Dictionary in the form:
    
    |  { <donor_element>: { <donor_charge>: { <receiver_element>: { <receiver_charge>: { <transition>: <pec> } } } } }, where
    |      <donor_element> is the electron donor species (Element/Isotope).
    |      <donor_charge> is the electron donor charge.
    |      <receiver_element> is the electron receiver species (Element/Isotope).
    |      <receiver_charge> is the electron receiver charge.
    |      <transition> is the tuple containing (initial level, final level).
    |      <pec> is the PEC dictionary containing the following entries:
    |          'ne': array-like of size (N) with electron density in m^-3,
    |          'te': array-like of size (M) with electron temperature in eV,
    |          'td': array-like of size (K) with donor temperature in eV,
    |          'rate': array-like of size (N, M, K) with PEC in photon.m^3.s^-1.
    
    :param repository_path: Path to the atomic data repository.
    """
    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    for donor_element, donor_charge_states in rates.items():
        for donor_charge, receiver_elements in donor_charge_states.items():
            for receiver_element, receiver_charge_states in receiver_elements.items():
                for receiver_charge, transitions in receiver_charge_states.items():

                    # sanitise and validate
                    if not isinstance(donor_element, Element):
                        raise TypeError('The donor element must be an Element object.')

                    if not valid_charge(donor_element, donor_charge + 1):
                        raise ValueError('Donor charge state is equal to or larger than the number of protons in the element.')

                    if not isinstance(receiver_element, Element):
                        raise TypeError('The receiver element must be an Element object.')

                    if not valid_charge(receiver_element, receiver_charge):
                        raise ValueError('Receiver charge state is larger than the number of protons in the element.')

                    rate_path = 'pec/thermal_cx/{0}/{1}/{2}/{3}.json'.format(donor_element.symbol.lower(), donor_charge,
                                                                             receiver_element.symbol.lower(), receiver_charge)
                    path = os.path.join(repository_path, rate_path)

                    # read in any existing rates
                    try:
                        with open(path, 'r') as f:
                            content = RecursiveDict.from_dict(json.load(f))
                    except FileNotFoundError:
                        content = RecursiveDict()

                    # add/replace data for a transition
                    for transition, data in transitions.items():
                        key = encode_transition(transition)

                        # sanitise/validate data
                        ne = np.array(data['ne'], np.float64)
                        te = np.array(data['te'], np.float64)
                        td = np.array(data['td'], np.float64)
                        rate = np.array(data['rate'], np.float64)

                        if ne.ndim != 1:
                            raise ValueError('Electron density array must be a 1D array.')

                        if te.ndim != 1:
                            raise ValueError('Electron temperature array must be a 1D array.')

                        if td.ndim != 1:
                            raise ValueError('Donor temperature array must be a 1D array.')

                        if (ne.shape[0], te.shape[0], td.shape[0]) != rate.shape:
                            raise ValueError('Electron density, electron temperature, donor temperature'
                                             ' and rate data arrays have inconsistent sizes.')

                        content[key] = {
                            'ne': ne.tolist(),
                            'te': te.tolist(),
                            'td': td.tolist(),
                            'rate': rate.tolist()
                        }

                    # create directory structure if missing
                    directory = os.path.dirname(path)
                    os.makedirs(directory, exist_ok=True)

                    # write new data
                    with open(path, 'w') as f:
                        json.dump(content.freeze(), f, indent=2, sort_keys=True)


def get_pec_excitation_rate(element, charge, transition, repository_path=None):
    """
    Reads the excitation PEC from the repository for the given
    element, charge and transition.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param transition: Tuple containing (initial level, final level).
    :param repository_path: Path to the atomic data repository.

    :return rate: Excitation PEC dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with excitation PEC in photon.m^3.s^-1.

    """

    return _get_pec_rate('excitation', element, charge, transition, repository_path)


def get_pec_recombination_rate(element, charge, transition, repository_path=None):
    """
    Reads the recombination PEC from the repository for the given
    element, charge and transition.

    :param element: Plasma species (Element/Isotope).
    :param charge: Charge of the plasma species.
    :param transition: Tuple containing (initial level, final level).
    :param repository_path: Path to the atomic data repository.

    :return rate: Recombination PEC dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'rate': 2D array of size (N, M) with recombination PEC in photon.m^3.s^-1.

    """

    return _get_pec_rate('recombination', element, charge, transition, repository_path)


def _get_pec_rate(cls, element, charge, transition, repository_path=None):

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH
    path = os.path.join(repository_path, 'pec/{}/{}/{}.json'.format(cls, element.symbol.lower(), charge))
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[encode_transition(transition)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested PEC rate (class={}, element={}, charge={}, transition={})'
                           ' is not available.'.format(cls, element.symbol, charge, transition))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d


def get_pec_thermal_cx_rate(donor_element, donor_charge, receiver_element, receiver_charge, transition, repository_path=None):
    """
    Reads the thermal charge exchange PEC from the repository for the given
    donor element, donor charge, receiver element, receiver charge and transition.

    :param donor_element: Electron donor plasma species (Element/Isotope).
    :param donor_charge: Electron donor charge.
    :param receiver_element: Electron receiver plasma species (Element/Isotope).
    :param receiver_charge: Electron receiver charge.
    :param transition: Tuple containing (initial level, final level).
    :param repository_path: Path to the atomic data repository.

    :return rate: Thermal CX PEC dictionary containing the following entries:

    |      'ne': 1D array of size (N) with electron density in m^-3,
    |      'te': 1D array of size (M) with electron temperature in eV,
    |      'td': 1D array of size (K) with donor temperature in eV,
    |      'rate': 2D array of size (N, M, K) with thermal CX PEC in photon.m^3.s^-1.

    """

    repository_path = repository_path or DEFAULT_REPOSITORY_PATH

    rate_path = 'pec/thermal_cx/{0}/{1}/{2}/{3}.json'.format(donor_element.symbol.lower(), donor_charge,
                                                             receiver_element.symbol.lower(), receiver_charge)
    path = os.path.join(repository_path, rate_path)
    try:
        with open(path, 'r') as f:
            content = json.load(f)
        d = content[encode_transition(transition)]
    except (FileNotFoundError, KeyError):
        raise RuntimeError('Requested thermal charge-exchange PEC (donor={}, donor charge={}, receiver={}, receiver charge={})'
                           ' is not available.'
                           ''.format(donor_element.symbol, donor_charge, receiver_element.symbol, receiver_charge))

    # convert to numpy arrays
    d['ne'] = np.array(d['ne'], np.float64)
    d['te'] = np.array(d['te'], np.float64)
    d['td'] = np.array(d['td'], np.float64)
    d['rate'] = np.array(d['rate'], np.float64)

    return d





cimport cython
from libc.math cimport exp, sqrt, pow
from scipy.constants import electron_mass, atomic_mass

from raysect.primitive import Box
from raysect.optical import World, Point3D, Vector3D

from cherab.core import Species, Maxwellian, Plasma
from cherab.core.math.function cimport Function3D
from cherab.core.atomic import hydrogen


cdef class NeutralFunction(Function3D):
    """A neutral profile that is constant outside the plasma, then exponentially decays
       inside the plasma boundary."""

    cdef double peak, sigma, pedestal_top
    cdef double _constant
    cdef bint _cache
    cdef double _cache_x, _cache_y, _cache_z, _cache_v

    def __init__(self, double peak_value, double sigma, double pedestal_top=1):

        self.peak = peak_value
        self.sigma = sigma
        self.pedestal_top = pedestal_top
        self._constant = (2*self.sigma*self.sigma)

        # last value cache
        self._cache = False
        self._cache_x = 0.0
        self._cache_y = 0.0
        self._cache_z = 0.0
        self._cache_v = 0.0

    @cython.cdivision(True)
    cdef double evaluate(self, double x, double y, double z) except? -1e999:

        cdef double value

        if self._cache:
            if x == self._cache_x and y == self._cache_y and z == self._cache_z:
                return self._cache_v

        if x >= 0:
            value = self.peak * exp(-(x**2) / self._constant)
        else:
            value = self.peak

        self._cache = True
        self._cache_x = x
        self._cache_y = y
        self._cache_z = z
        self._cache_v = value

        return value


cdef class IonFunction(Function3D):
    """An approximate pedestal plasma profile that follows a double
       quadratic between the plasma boundary and the pedestal top."""

    cdef double t_core, t_lcfs, pedestal_top, p, q
    cdef bint _cache
    cdef double _cache_x, _cache_y, _cache_z, _cache_v

    def __init__(self, double t_core, double t_lcfs, double p=2, double q=2, double pedestal_top=1):

        self.t_core = t_core
        self.t_lcfs = t_lcfs
        self.p = p
        self.q = q
        self.pedestal_top = pedestal_top

        # last value cache
        self._cache = False
        self._cache_x = 0.0
        self._cache_y = 0.0
        self._cache_z = 0.0
        self._cache_v = 0.0

    @cython.cdivision(True)
    cdef double evaluate(self, double x, double y, double z) except? -1e999:

        cdef double value, x_norm

        if self._cache:
            if x == self._cache_x and y == self._cache_y and z == self._cache_z:
                return self._cache_v

        x_norm = x / self.pedestal_top
        if 0 <= x_norm <= 1:
            value = ((self.t_core - self.t_lcfs) *
                    pow((1 - pow((1 - x_norm), self.p)), self.q) + self.t_lcfs)
        elif x_norm >= 1:
            value = self.t_core
        else:
            value = 0.0

        self._cache = True
        self._cache_x = x
        self._cache_y = y
        self._cache_z = z
        self._cache_v = value

        return value


# TODO - replace with ionisation balance calculations
def build_slab_plasma(length=5, width=1, height=1, peak_density=1e19, peak_temperature=2500,
                      pedestal_top=1, neutral_temperature=0.5, impurities=None,
                      parent=None):
    """
    Constructs a simple slab of plasma.

    The plasma is defined for positive x starting at x = 0, symmetric in y-z. The plasma
    parameters such as electron density and temperature evolve in 1 dimension according
    to the input parameters specified. The slab includes an optional pedestal.

    Raysect cannot handle infinite geometry, so overall spatial dimensions of the slab need
    to be set, [length, width, height]. These can be set very large to make an effectively
    infinite slab of plasma, although the numerical performance will degrade accordingly.
    The dimensions should be set appropriately with valid assumptions for your scenario.

    Impurity species can be included as a list of tuples, where each tuple specifies an
    impurity. The specification format is (species, charge, concentration). For example:

        >>> impurities=[(carbon, 6, 0.005)]

    :param float length: the overall length of the slab along x.
    :param float width: the y width of the slab.
    :param float height: the z height of the slab.
    :param float peak_density: the peak electron density at the pedestal top.
    :param float peak_temperature: the peak electron temperature at the pedestal top.
    :param float pedestal_top: the length of the pedestal top.
    :param float neutral_temperature: the background neutral temperature.
    :param list impurities: an optional list of impurities to include.
    :param parent: the Raysect scene-graph parent node.
    :param atomic_data: the atomic data provider to use for subsequent spectroscopic calculations,
      defaults to atomic_data=OpenADAS(permit_extrapolation=True).

    .. code-block:: pycon

       >>> from raysect.optical import World
       >>> from cherab.core.atomic import carbon
       >>> from cherab.tools.plasmas.slab import build_slab_plasma
       >>>
       >>> plasma = build_slab_plasma(peak_density=5e19, impurities=[(carbon, 6, 0.005)])
       >>> plasma.parent = World()
    """

    plasma = Plasma(parent=parent)
    plasma.geometry = Box(Point3D(0, -width/2, -height/2), Point3D(length, width/2, height/2))

    species = []

    # No net velocity for any species
    zero_velocity = Vector3D(0, 0, 0)

    # define neutral species distribution
    h0_density = NeutralFunction(peak_density, 0.1, pedestal_top=pedestal_top)
    h0_temperature = neutral_temperature
    h0_distribution = Maxwellian(h0_density, h0_temperature, zero_velocity,
                                 hydrogen.atomic_weight * atomic_mass)
    species.append(Species(hydrogen, 0, h0_distribution))

    # define hydrogen ion species distribution
    h1_density = IonFunction(peak_density, 0, pedestal_top=pedestal_top)
    h1_temperature = IonFunction(peak_temperature, 0, pedestal_top=pedestal_top)
    h1_distribution = Maxwellian(h1_density, h1_temperature, zero_velocity,
                                 hydrogen.atomic_weight * atomic_mass)
    species.append(Species(hydrogen, 1, h1_distribution))

    # add impurities
    if impurities:
        for impurity, ionisation, concentration in impurities:
            imp_density = IonFunction(peak_density * concentration, 0, pedestal_top=pedestal_top)
            imp_temperature = IonFunction(peak_temperature, 0, pedestal_top=pedestal_top)
            imp_distribution = Maxwellian(imp_density, imp_temperature, zero_velocity,
                                          impurity.atomic_weight * atomic_mass)
            species.append(Species(impurity, ionisation, imp_distribution))

    # define the electron distribution
    e_density = IonFunction(peak_density, 0, pedestal_top=pedestal_top)
    e_temperature = IonFunction(peak_temperature, 0, pedestal_top=pedestal_top)
    e_distribution = Maxwellian(e_density, e_temperature, zero_velocity, electron_mass)

    # define species
    plasma.b_field = Vector3D(0, 0, 0)
    plasma.electron_distribution = e_distribution
    plasma.composition = species

    return plasma


def build_constant_slab_plasma(length=5, width=1, height=1, electron_density=1e19, electron_temperature=2.5e3,
                               plasma_species=None, b_field=Vector3D(0, 0, 0), parent=None):
    """
    Constructs a simple slab of plasma with constant conditions.

    The plasma is defined for positive x starting at x = 0, symmetric in y-z. The plasma
    parameters such as electron density and temperature are constant over the plasma volume.

    Raysect cannot handle infinite geometry, so overall spatial dimensions of the slab need
    to be set, [length, width, height]. These can be set very large to make an effectively
    infinite slab of plasma, although the numerical performance will degrade accordingly.
    The dimensions should be set appropriately with valid assumptions for your scenario.

    Ion species can be included as a list of tuples, where each tuple specifies an
    impurity. The specification format is (species, charge, density, temperature, velocity). For example:

        >>> plasma_species = [(carbon, 6, 1e18, 3.4e3, Vector3D(1.0e3, 0, 0))]

    If omitted, hydrogen distribution with properties equal to electrons is used:
        >>> plasma_species = [(hydrogen, 1, electron_density, electron_temperature, Vector3D(0, 0, 0))]
    
    If an empty list is passed, plasma contains only electrons.

    :param float length: the overall length of the slab along x.
    :param float width: the y width of the slab.
    :param float height: the z height of the slab.
    :param float electron_density: the electron density in m^-3 .
    :param float electron_temperature: the electron temperature in eV.
    :param list plasma_species: an optional list of impurities to include.
    :param Vector3D b_field: vector giving the magnetic field
    :param parent: the Raysect scene-graph parent node.

    .. code-block:: pycon

       >>> from raysect.optical import World
       >>> from cherab.core.atomic import hydrogen, carbon
       >>> from cherab.tools.plasmas.slab import build_constant_slab_plasma
       >>>
       >>> plasma_species = [(hydrogen, 0, 1e19, 3.5e3, Vector3D(5e3, 0, 0)), (carbon, 5, 1e18, 3.4e3, Vector3D(1.0e3, 0, 0))]
       >>> plasma = build_constant_slab_plasma(0.2, 0.5, 0.5, electron_density = 1.19, electron_temperature=4e4, plasma_species=plasma_species, b_field=Vector3D(0, 5, 0))
       >>> plasma.parent = World()
    """

    if plasma_species is None:
        plasma_species = [(hydrogen, 1, electron_density, electron_temperature, Vector3D(0, 0, 0))]

    # create electron distribution
    e_distribution = Maxwellian(electron_density, electron_temperature, Vector3D(0, 0, 0), electron_mass)

    # create ion species
    species = []
    for element, ionisation, density, temperature, velocity in plasma_species:
        imp_distribution = Maxwellian(density, temperature, velocity, element.atomic_weight * atomic_mass)
        species.append(Species(element, ionisation, imp_distribution))

    # create plasma, add particles and magnetic field
    plasma = Plasma(parent=parent)
    plasma.geometry = Box(Point3D(0, -width/2, -height/2), Point3D(length, width/2, height/2))
    plasma.electron_distribution = e_distribution
    plasma.b_field = b_field
    plasma.composition = species

    return plasma



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


# External imports
import os
from scipy.constants import electron_mass, atomic_mass

import matplotlib.pyplot as plt
import numpy as np
from cherab.core.model import ExcitationLine, RecombinationLine, Bremsstrahlung

# Cherab and raysect imports
from cherab.core import Species, Maxwellian, Plasma, Line, elements
from cherab.openadas import OpenADAS
from cherab.tools.plasmas import GaussianVolume

# Core and external imports
from raysect.optical import World, translate, rotate, Vector3D, Point3D, Ray
from raysect.primitive import Sphere
from raysect.optical.observer import PinholeCamera
from raysect.optical.material.emitter.inhomogeneous import NumericalIntegrator

# tunables
ion_density = 1e19
sigma = 0.25

# setup scenegraph
world = World()

# create atomic data source
adas = OpenADAS(permit_extrapolation=True)

# PLASMA ----------------------------------------------------------------------
plasma = Plasma(parent=world)
plasma.atomic_data = adas
plasma.geometry = Sphere(sigma * 5.0)
plasma.geometry_transform = None
plasma.integrator = NumericalIntegrator(step=sigma / 5.0)

# define basic distributions
d_density = GaussianVolume(0.5 * ion_density, sigma*10000)
e_density = GaussianVolume(ion_density, sigma*10000)
temperature = 1 + GaussianVolume(79, sigma)
bulk_velocity = Vector3D(-1e5, 0, 0)

d_mass = elements.deuterium.atomic_weight * atomic_mass
d_distribution = Maxwellian(d_density, temperature, bulk_velocity, d_mass)
e_distribution = Maxwellian(e_density, temperature, bulk_velocity, electron_mass)

d0_species = Species(elements.deuterium, 0, d_distribution)
d1_species = Species(elements.deuterium, 1, d_distribution)

# define species
plasma.b_field = Vector3D(1.0, 1.0, 1.0)
plasma.electron_distribution = e_distribution
plasma.composition = [d0_species, d1_species]

# Setup elements.deuterium lines
d_alpha = Line(elements.deuterium, 0, (3, 2))
d_beta = Line(elements.deuterium, 0, (4, 2))
d_gamma = Line(elements.deuterium, 0, (5, 2))
d_delta = Line(elements.deuterium, 0, (6, 2))
d_epsilon = Line(elements.deuterium, 0, (7, 2))

plasma.models = [
    Bremsstrahlung(),
    ExcitationLine(d_alpha),
    ExcitationLine(d_beta),
    ExcitationLine(d_gamma),
    ExcitationLine(d_delta),
    ExcitationLine(d_epsilon),
    RecombinationLine(d_alpha),
    RecombinationLine(d_beta),
    RecombinationLine(d_gamma),
    RecombinationLine(d_delta),
    RecombinationLine(d_epsilon)
]


plt.ion()

r = Ray(origin=Point3D(0, 0, -5), direction=Vector3D(0, 0, 1),
        min_wavelength=100, max_wavelength=1000, bins=1e6)
s = r.trace(world)
plt.plot(s.wavelengths, s.samples)

r = Ray(origin=Point3D(-5, 0, -5), direction=Vector3D(1, 0, 1),
        min_wavelength=100, max_wavelength=1000, bins=1e6)
s = r.trace(world)
plt.plot(s.wavelengths, s.samples)

r = Ray(origin=Point3D(-5, 0, 0), direction=Vector3D(1, 0, 0),
        min_wavelength=100, max_wavelength=1000, bins=1e6)
s = r.trace(world)
plt.plot(s.wavelengths, s.samples)

plt.xlabel('Wavelength (nm)')
plt.ylabel('Radiance (W/m^2/str/nm)')
plt.title('Sampled Balmer Series Spectrum')
plt.show()

camera = PinholeCamera((128, 128), parent=world, transform=translate(0, 0, -3.5))
camera.spectral_rays = 1
camera.spectral_bins = 15
camera.pixel_samples = 50

plt.ion()
camera.observe()

plt.ioff()
plt.show()



# cython: language_level=3

# Copyright 2016-2023 Euratom
# Copyright 2016-2023 United Kingdom Atomic Energy Authority
# Copyright 2016-2023 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


cdef class BeamLineShapeModel:
    """
    A base class for building beam emission line shapes.

    :param Line line: The emission line object for this line shape.
    :param float wavelength: The rest wavelength for this emission line.
    :param Beam beam: The beam class that is emitting.
    :param AtomicData atomic_data: The atomic data provider.
    """

    def __init__(self, Line line, double wavelength, Beam beam, AtomicData atomic_data):

        self.line = line
        self.wavelength = wavelength
        self.beam = beam
        self.atomic_data = atomic_data

    cpdef Spectrum add_line(self, double radiance, Point3D beam_point, Point3D plasma_point,
                            Vector3D beam_direction, Vector3D observation_direction, Spectrum spectrum):
        raise NotImplementedError('Child lineshape class must implement this method.')



import unittest

from raysect.optical import World
from raysect.optical.material.emitter.inhomogeneous import NumericalIntegrator

from cherab.core import Plasma
from cherab.core.laser.node import Laser
from cherab.core.model.laser.laserspectrum import ConstantSpectrum
from cherab.core.model.laser.model import SeldenMatobaThomsonSpectrum
from cherab.core.model.laser.profile import UniformEnergyDensity


class TestLaser(unittest.TestCase):

    def test_laser_init(self):
        """
        Test correct initialisation of a laser instance.
        """

        world = World()
        laser = Laser(parent=world)

        with self.assertRaises(ValueError, msg="Model was attached before Plasma, Profile and LaserSpectrum were specified."):
            laser.models = [SeldenMatobaThomsonSpectrum()]

        laser.laser_profile = UniformEnergyDensity()
        with self.assertRaises(ValueError, msg="Model was attached before Plasma, Profile and LaserSpectrum were specified."):
            laser.models = [SeldenMatobaThomsonSpectrum()]

        laser.laser_spectrum = ConstantSpectrum(min_wavelength=1059, max_wavelength=1061, bins=10)
        with self.assertRaises(ValueError, msg="Model was attached before Plasma, Profile and LaserSpectrum were specified."):
            laser.models = [SeldenMatobaThomsonSpectrum()]

        laser.plasma = Plasma(parent=world)
        laser.models = [SeldenMatobaThomsonSpectrum()]

    def test_reference_change(self):

        world = World()

        laser_profile = UniformEnergyDensity(laser_length=1, laser_radius=0.1)
        laser_spectrum = ConstantSpectrum(min_wavelength=1059, max_wavelength=1061, bins=10)
        plasma = Plasma(parent=world)
        models = [SeldenMatobaThomsonSpectrum()]

        laser_profile2 = UniformEnergyDensity()
        laser_spectrum2 = ConstantSpectrum(min_wavelength=1059, max_wavelength=1061, bins=10)
        plasma2 = Plasma(parent=world)
        models2 = [SeldenMatobaThomsonSpectrum()]

        laser = Laser(parent=world)

        laser.laser_spectrum = laser_spectrum
        laser.plasma = plasma
        laser.laser_profile = laser_profile
        laser.models = models

        for mod in list(laser.models):
            self.assertIs(mod.laser_profile, laser_profile, msg="laser_profile reference in emission model"
                          "is not set correctly.")
            self.assertIs(mod.plasma, plasma, msg="plasma reference in emission model"
                                                            "is not set correctly.")
            self.assertIs(mod.laser_spectrum, laser_spectrum, msg="laser_spectrum reference in emission model"
                                                            "is not set correctly.")

        laser.laser_spectrum = laser_spectrum2
        laser.plasma = plasma2
        laser.laser_profile = laser_profile2

        for mod in list(laser.models):
            self.assertIs(mod.laser_profile, laser_profile2, msg="laser_profile reference in emission model"
                                                            "is not set correctly.")
            self.assertIs(mod.plasma, plasma2, msg="plasma reference in emission model"
                                                  "is not set correctly.")
            self.assertIs(mod.laser_spectrum, laser_spectrum2, msg="laser_spectrum reference in emission model"
                                                                  "is not set correctly.")

        laser.models = models + models2

        for mod in list(laser.models):
            self.assertIs(mod.laser_profile, laser_profile2, msg="laser_profile reference in emission model"
                                                             "is not set correctly.")
            self.assertIs(mod.plasma, plasma2, msg="plasma reference in emission model"
                                                   "is not set correctly.")
            self.assertIs(mod.laser_spectrum, laser_spectrum2, msg="laser_spectrum reference in emission model"
                                                                   "is not set correctly.")

    def test_integrator_change(self):

        world = World()

        laser_profile = UniformEnergyDensity(laser_length=1, laser_radius=0.1)
        laser_spectrum = ConstantSpectrum(min_wavelength=1059, max_wavelength=1061, bins=10)
        plasma = Plasma(parent=world)
        models = [SeldenMatobaThomsonSpectrum()]

        laser = Laser(parent=world)

        laser.laser_spectrum = laser_spectrum
        laser.plasma = plasma
        laser.laser_profile = laser_profile
        laser.models = models

        integrator = NumericalIntegrator(1e-4)

        laser.integrator = integrator

        for i in laser.get_geometry():
            self.assertIs(i.material.integrator, integrator, msg="Integrator not updated properly")
             



"""
This example demonstrates creating a radiating object with radiation
defined as a 3D function f(x, y, z), and observing that radiation with
the bolometers. Note that in this case the 3D radiation profile is
produced by defining an axisymmetric 2D profile and mapping that to 3D.

5 measurements are presented:
- Power measured on the foil
- Radiance measured by a single line of sight observer
- Radiance measured on the foil, using the volumetric power divided by
  the etendue
- Radiance measured on the foil, using the radiance pipeline
- Brightness measured on the foil, assuming the power is radiated
  isotropically

Note that the radiance pipeline measures the radiance averaged over the
entire solid angle of the detector, which is 2pi steradians for a
bolometer foil. To get a meaningful comparison with the sightline
observer we need to renormalise to the fractional solid angle subtended
by the foil-slit system.
"""
import math
import matplotlib.pyplot as plt

from raysect.core import Node, Point2D, Point3D, Vector3D, rotate_basis, translate
from raysect.optical import World
from raysect.optical.material import AbsorbingSurface, VolumeTransform
from raysect.primitive import Box, Cylinder, Subtract

from cherab.core.math import AxisymmetricMapper, sample2d
from cherab.tools.emitters import RadiationFunction
from cherab.tools.observers.bolometry import BolometerCamera, BolometerSlit, BolometerFoil


# Convenient constants
XAXIS = Vector3D(1, 0, 0)
YAXIS = Vector3D(0, 1, 0)
ZAXIS = Vector3D(0, 0, 1)
ORIGIN = Point3D(0, 0, 0)
# Bolometer geometry
BOX_WIDTH = 0.05
BOX_HEIGHT = 0.07
BOX_DEPTH = 0.2
SLIT_WIDTH = 0.004
SLIT_HEIGHT = 0.005
FOIL_WIDTH = 0.0013
FOIL_HEIGHT = 0.0038
FOIL_CORNER_CURVATURE = 0.0005
SLIT_SENSOR_SEPARATION = 0.1
FOIL_SEPARATION = 0.00508  # 0.2 inch between foils

world = World()

########################################################################
# Build a simple bolometer camera.
########################################################################

# The camera consists of a box with a rectangular slit and 4 foils.
# In its local coordinate system, the camera's slit is located at the
# origin and the foils below the z=0 plane, looking up towards the slit.
camera_box = Box(lower=Point3D(-BOX_WIDTH / 2, -BOX_HEIGHT / 2, -BOX_DEPTH),
                 upper=Point3D(BOX_WIDTH / 2, BOX_HEIGHT / 2, 0))
# Hollow out the box
inside_box = Box(lower=camera_box.lower + Vector3D(1e-5, 1e-5, 1e-5),
                 upper=camera_box.upper - Vector3D(1e-5, 1e-5, 1e-5))
camera_box = Subtract(camera_box, inside_box)
# The slit is a hole in the box
aperture = Box(lower=Point3D(-SLIT_WIDTH / 2, -SLIT_HEIGHT / 2, -1e-4),
               upper=Point3D(SLIT_WIDTH / 2, SLIT_HEIGHT / 2, 1e-4))
camera_box = Subtract(camera_box, aperture)
camera_box.material = AbsorbingSurface()
# Instance of the bolometer camera
bolometer_camera = BolometerCamera(camera_geometry=camera_box)
# The bolometer slit in this instance just contains targeting information
# for the ray tracing, since we have already given our camera a geometry
# The slit is defined in the local coordinate system of the camera
slit = BolometerSlit(slit_id="Example slit", centre_point=ORIGIN,
                     basis_x=XAXIS, dx=SLIT_WIDTH, basis_y=YAXIS, dy=SLIT_HEIGHT,
                     parent=bolometer_camera, csg_aperture=False)
# 4 bolometer foils, spaced at equal intervals along the local X axis
# The bolometer positions and orientations are given in the local coordinate
# system of the camera, just like the slit
sensor = Node(name="Bolometer sensor", parent=bolometer_camera,
              transform=translate(0, 0, -SLIT_SENSOR_SEPARATION))
# The foils are shifted relative to the centre of the sensor by -1.5, -0.5, 0.5 and 1.5
# times the foil-foil separation
for i, shift in enumerate([-1.5, -0.5, 0.5, 1.5]):
    foil_transform = translate(shift * FOIL_SEPARATION, 0, 0) * sensor.transform
    foil = BolometerFoil(detector_id="Foil {}".format(i + 1),
                         centre_point=ORIGIN.transform(foil_transform),
                         basis_x=XAXIS.transform(foil_transform), dx=FOIL_WIDTH,
                         basis_y=YAXIS.transform(foil_transform), dy=FOIL_HEIGHT,
                         slit=slit, parent=bolometer_camera, units="Power",
                         accumulate=False, curvature_radius=FOIL_CORNER_CURVATURE)
    bolometer_camera.add_foil_detector(foil)

bolometer_camera.transform = translate(1, 0, 1.5) * rotate_basis(-ZAXIS, YAXIS)
bolometer_camera.parent = world

########################################################################
# Produce a simple radiating plasma.
########################################################################

# The plasma will be a cylindrical plasma which emits with a constant
# emissivity of 1 W/m3 in an annular ring.
# See the RadiationFunction example for more details of how this is set up.
MAJOR_RADIUS = 1
MINOR_RADIUS = 0.5
CENTRE_Z = -0.5
CYLINDER_HEIGHT = 5
EMISSIVITY = 1

def emission_function(r, z):
    if (r - MAJOR_RADIUS)**2 + (z - CENTRE_Z)**2 < MINOR_RADIUS**2:
        return EMISSIVITY
    return 0

emitter = Cylinder(radius=MAJOR_RADIUS + MINOR_RADIUS, height=CYLINDER_HEIGHT,
                   transform=translate(0, 0, CENTRE_Z - CYLINDER_HEIGHT / 2))
emission_function_3d = AxisymmetricMapper(emission_function)
emitting_material = VolumeTransform(RadiationFunction(emission_function_3d),
                                    transform=emitter.transform.inverse())
emitter.material = emitting_material

########################################################################
# Plot the bolometer lines of sight and the radiation function
########################################################################
floor = Box(lower=Point3D(-10, -10, -1.26), upper=Point3D(10, 10, -1.25), parent=world,
            material=AbsorbingSurface(), name="Z=-1.25 plane")


def _point3d_to_rz(point):
    return Point2D(math.hypot(point.x, point.y), point.z)


fig, ax = plt.subplots()
for foil in bolometer_camera:
    slit_centre = foil.slit.centre_point
    slit_centre_rz = _point3d_to_rz(slit_centre)
    ax.plot(slit_centre_rz[0], slit_centre_rz[1], 'ko')
    origin, hit, _ = foil.trace_sightline()
    centre_rz = _point3d_to_rz(foil.centre_point)
    ax.plot(centre_rz[0], centre_rz[1], 'kx')
    origin_rz = _point3d_to_rz(origin)
    hit_rz = _point3d_to_rz(hit)
    ax.plot([origin_rz[0], hit_rz[0]], [origin_rz[1], hit_rz[1]], 'k')

r, z, emiss_sampled = sample2d(
    emission_function, (0.25, 1.75, 150), (-1.25, 0.25, 150)
)
image = ax.imshow(emiss_sampled.T, origin="lower", zorder=-10,
                  extent=(r.min(), r.max(), z.min(), z.max()))
fig.colorbar(image)
ax.set_xlabel("r")
ax.set_ylabel("z")


########################################################################
# Measure the radiation with the bolometers
########################################################################

emitter.parent = world

for foil in bolometer_camera:
    # Ensure reasonable sampling statistics
    foil.pixel_samples = 100000
    # Measure the incident power
    foil.units = "Power"
    foil.observe()
    power = foil.pipelines[0].value.mean
    power_error = foil.pipelines[0].value.error()
    print("Measured power for {}:\t\t{:.03g} +- {:.1g} W".format(foil.name, power, power_error))
    # Measure the incident radiance with a sightline
    foil.units = "Radiance"
    sightline = foil.as_sightline()
    sightline.observe()
    sightline_radiance = sightline.pipelines[0].value.mean
    print("Sightline radiance for {}:\t\t{:.03g} W/m2/sr".format(sightline.name, sightline_radiance))
    # Calculate the incident radiance from the power
    emitter.parent = None  # No other objects should be in the scene for etendue calculation
    floor.parent = None
    etendue, etendue_error = foil.calculate_etendue()
    radiance_from_power = power / etendue
    radiance_from_power_error = (math.hypot(power_error / power, etendue_error / etendue)
                                 * radiance_from_power)
    print("Radiance from power for {}:\t\t{:.03g} +- {:.1g} W/m2/sr".format(
        foil.name, radiance_from_power, radiance_from_power_error))
    # Calculate the incident radiance directly, renormalising for comparison with the sightline
    emitter.parent = world
    foil.units = "Radiance"
    foil.observe()
    fractional_solid_angle = etendue / foil.sensitivity
    radiance = foil.pipelines[0].value.mean / fractional_solid_angle
    radiance_error = (
        math.hypot(foil.pipelines[0].value.error() / foil.pipelines[0].value.mean,
                   etendue_error / etendue)
        * radiance
    )
    print("Measured radiance for {}:\t\t{:.03g} +- {:.1g} W/m2/sr".format(
        foil.name, radiance, radiance_error))
    # Calculate the brightness, assuming power is radiated isotropically in 4pi steradians
    brightness = radiance * 4 * math.pi
    brightness_error = radiance_error / radiance * brightness
    print("Calculated brightness for {}:\t{:.03g} +- {:.1g} W/m2".format(
        foil.name, brightness, brightness_error))
    print()

plt.show()



# -*- coding: utf-8 -*-
#
# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import os
import numpy as np
from timeit import default_timer as timer
from .opencl_utils import device_select
try:
    import pyopencl as cl
except ImportError:
    _has_pyopencl = False
else:
    _has_pyopencl = True


class SartOpencl:
    """
    A GPU-accelerated version of SART inversion.
    The geometry matrix and Laplacian matrix are provided on initialisation because they must
    be copied to GPU memory, which takes time. Inversions may be performed multiple times
    for different measurement vectors without copying the matrices each time. If required,
    the Laplacian matrix can be updated by calling
    `update_laplacian_matrix(new_laplacian_matrix)` method. Note that computations are
    performed with single precision.

    :param np.ndarray geometry_matrix: The sensitivity matrix describing the coupling between
                                       the detectors and the voxels. Must be an array
                                       with shape :math:`(N_d, N_s)`.
    :param np.ndarray laplacian_matrix:  The laplacian regularisation matrix of
                                         shape :math:`(N_s, N_s)`.
                                         Default value: `laplacian_matrix=None`.
    :param pyopencl.Device device: OpenCL device which will be used for computations.
                                   Default value: `device=None` (autoselect).
    :param int block_size: Number of GPU threads per block. Must be the power of 2.
                           For the best performance try from 256 to 1024 for Nvidia
                           (use 1024 on high-end GPUs), from 64 to 256 for AMD and from 16 to 64
                           for Intel GPUs. Default value: `block_size=256`.
    :param bool copy_column_major: If True, the two copies of geometry matrix will be stored in
                                   GPU memory. One in row-major order and the other one in
                                   column-major order. This provides much better performance
                                   of the inversions but requires twice as much GPU memory.
                                   Default value: `copy_column_major=True`.
    :param int block_size_row_maj: If `copy_column_major` is set to False, this parameter defines
                                   the number of GPU threads per block in mat_vec_mult_row_maj()
                                   kernel used to calculate y_hat. Must be lower than
                                   `block_size`. Default value: `block_size_row_maj=64` (optimal
                                   value for Nvidia GPUs).
    :param bool use_atomic: If True, increases the number of thread blocks that can run in
                            parallel with the help of atomic operations (custom atomic add on floats).
                            Set this to False, if the atomic operations are running slow on your device
                            (Nvidia GPUs before Kepler, some AMD APUs, some Intel GPUs).
                            Default value: `use_atomic=True`.
    :param int steps_per_thread: If `use_atomic` is set to True, this parameters defines the
                                 maximum number of loop steps performed by the parallel threads
                                 in a single thread block. Default value: `steps_per_thread=64`
                                 (optimal for Nvidia GPUs).
    :param bool verbose: Verbose output, defaults to `verbose=False`.

    .. code-block:: pycon

        >>> with SartOpencl(geometry_matrix, block_size=1024) as invert_sart:
        >>>     solution, residual = invert_sart(measurement_vector)
        >>> ### or ###
        >>> inv_sart = SartOpencl(geometry_matrix, block_size=1024)
        >>> solution, residual = inv_sart(measurement_vector)
        >>> inv_sart.clean()
    """

    def __init__(self, geometry_matrix, laplacian_matrix=None, device=None, block_size=256, copy_column_major=True, block_size_row_maj=64,
                 use_atomic=True, steps_per_thread=64, verbose=False):
        if not _has_pyopencl:
            raise RuntimeError("The pyopencl module is required to use the SartOpencl() inversion class.")

        self._verbose = verbose

        if geometry_matrix.dtype != np.float32:  # converting geometry_matrix to float32 if needed
            geometry_matrix = geometry_matrix.astype(np.float32)

        self._m_detectors, self._n_sources = geometry_matrix.shape
        cell_ray_densities = geometry_matrix.sum(0)
        ray_lengths = geometry_matrix.sum(1)
        device = device or device_select(verbose=verbose)
        self._use_atomic = use_atomic
        steps_per_thread = min(block_size, steps_per_thread)
        steps_per_thread_row_maj = min(block_size_row_maj, steps_per_thread)

        # creating OpenCL context
        self._cl_context = cl.Context([device])

        # reading and compiling OpenCL kernels
        kernels_filename = 'sart_kernels_atomic.cl' if use_atomic else 'sart_kernels.cl'
        kernel_source_file = os.path.join(os.path.dirname(os.path.realpath(__file__)), kernels_filename)
        with open(kernel_source_file) as f_kernel:
            kernel_source = f_kernel.read()
        compile_options = ['-DBLOCK_SIZE={}'.format(block_size), '-DSTEPS_PER_THREAD={}'.format(steps_per_thread),
                           '-DSTEPS_PER_THREAD_ROW_MAJ={}'.format(steps_per_thread_row_maj),
                           '-DBLOCK_SIZE_ROW_MAJ={}'.format(block_size_row_maj), '-cl-fast-relaxed-math']
        self._cl_prog = cl.Program(self._cl_context, kernel_source).build(options=compile_options)

        # creating buffers in device memory
        mf = cl.mem_flags

        self._geometry_matrix_device = cl.Buffer(self._cl_context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=geometry_matrix)
        if copy_column_major:
            geometry_matric_col_maj = geometry_matrix.flatten(order='F')
            self._geometry_matric_col_maj_device = cl.Buffer(self._cl_context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=geometry_matric_col_maj)
        else:
            self._geometry_matric_col_maj_device = None

        if laplacian_matrix is not None:
            laplacian_matrix = laplacian_matrix.flatten(order='F').astype(np.float32)
            self._laplacian_matrix_device = cl.Buffer(self._cl_context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=laplacian_matrix)
        else:
            self._laplacian_matrix_device = None

        self._cell_ray_densities_device = cl.Buffer(self._cl_context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=cell_ray_densities)
        self._ray_lengths_device = cl.Buffer(self._cl_context, mf.READ_ONLY | mf.COPY_HOST_PTR, hostbuf=ray_lengths)
        grad_penalty = np.zeros(self._n_sources, dtype=np.float32)
        self._grad_penalty_device = cl.Buffer(self._cl_context, mf.READ_WRITE | mf.COPY_HOST_PTR, hostbuf=grad_penalty)
        self._solution_device = cl.Buffer(self._cl_context, mf.READ_WRITE, cell_ray_densities.nbytes)
        self._detectors_device = cl.Buffer(self._cl_context, mf.READ_ONLY, ray_lengths.nbytes)
        self._y_hat_device = cl.Buffer(self._cl_context, mf.READ_WRITE, ray_lengths.nbytes)

        # calculating global and local work sizes
        nrem = self._n_sources % block_size
        gws_sources_x = self._n_sources + bool(nrem) * (block_size - nrem)
        mrem = self._m_detectors % block_size
        gws_detectors_x = self._m_detectors + bool(mrem) * (block_size - mrem)
        mrem_rm = self._m_detectors % block_size_row_maj
        gws_detectors_row_maj_x = self._m_detectors + bool(mrem_rm) * (block_size - mrem_rm)
        if use_atomic:
            gws_sources_row_maj_y = self._n_sources // steps_per_thread_row_maj + bool(self._n_sources % steps_per_thread_row_maj)
            gws_sources_y = self._n_sources // steps_per_thread + bool(self._n_sources % steps_per_thread)
            gws_detectors_y = self._m_detectors // steps_per_thread + bool(self._m_detectors % steps_per_thread)
        else:
            gws_sources_row_maj_y = gws_sources_y = gws_detectors_y = 1
        self._global_work_size = {}
        self._local_work_size = {'default': (block_size, 1)}
        self._global_work_size['trivial_sources'] = (gws_sources_x, 1)
        self._global_work_size['trivial_detectors'] = (gws_detectors_x, 1)
        self._global_work_size['iter'] = (gws_sources_x, gws_detectors_y)
        if copy_column_major:
            self._local_work_size['mult'] = self._local_work_size['default']
            self._global_work_size['mult'] = (gws_detectors_x, gws_sources_y)
        else:
            self._local_work_size['mult'] = (block_size_row_maj, 1)
            self._global_work_size['mult'] = (gws_detectors_row_maj_x, gws_sources_row_maj_y)
        self._global_work_size['grad'] = (gws_sources_x, gws_sources_y)

    def clean(self):
        """ Releases GPU buffers"""
        self._geometry_matrix_device.release()
        if self._geometry_matric_col_maj_device is not None:
            self._geometry_matric_col_maj_device.release()
        if self._laplacian_matrix_device is not None:
            self._laplacian_matrix_device.release()
        self._cell_ray_densities_device.release()
        self._ray_lengths_device.release()
        self._solution_device.release()
        self._grad_penalty_device.release()
        self._detectors_device.release()
        self._y_hat_device.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.clean()

    def update_laplacian_matrix(self, laplacian_matrix):
        """
        Updates the Laplacian matrix in GPU memory

        :param np.ndarray laplacian_matrix:  The laplacian regularisation matrix of
            shape :math:`(N_s, N_s)`.
        """
        if self._laplacian_matrix_device is not None:
            laplacian_matrix = laplacian_matrix.flatten(order='F').astype(np.float32)
            queue = cl.CommandQueue(self._cl_context)
            cl.enqueue_copy(queue, self._laplacian_matrix_device, laplacian_matrix)

    def __call__(self, measurement_vector, initial_guess=None, max_iterations=250, relaxation=1.0,
                 beta_laplace=0.01, conv_tol=1.e-4, time_limit=None):
        r"""
        Performs the inversion for a given measurement vector.

        :param np.ndarray measurement_vector: The measured power/radiance vector with
            shape :math:`(N_d)`.
        :param initial_guess: An optional initial guess, can be an array of shape :math:`(N_s)`
            or a constant value that will be used to seed the algorithm.
        :param int max_iterations: The maximum number of iterations to run the SART algorithm
            before returning a result, defaults to `max_iterations=250`.
        :param float relaxation: The relaxation hyperparameter, defaults to `relaxation=1`.
            Consult the reference papers for more information on this hyperparameter.
        :param float beta_laplace: The regularisation hyperparameter in the range [0, 1].
            Defaults to `beta_laplace=0.01`.
        :param float conv_tol: The convergence limit at which the algorithm will be terminated,
            unless the maximum number of iterations has been reached. The convergence is
            calculated as the normalised squared difference between the measurement and solution
            vectors. Note that reaching convergence lower than 1.e-6 is hardly possible
            on GPUs due to single precision calculations and relaxed math.
        :param float time_limit: If set, the iterations will stop after this time limit (in
            seconds) is reached. Default value: `time_limit=None`.

        :return: A tuple with the inverted solution vector :math:`\mathbf{x}` as an ndarray with
            shape :math:`(N_s)`, and the list of convergence values achieved after each iteration
            step.
        """

        time_start = timer()
        time_limit = time_limit or 1.e7
        if initial_guess is None:
            solution = np.zeros(self._n_sources, dtype=np.float32) + 1 / np.e
        elif isinstance(initial_guess, (float, int)):
            solution = np.zeros(self._n_sources, dtype=np.float32) + initial_guess
        else:
            solution = initial_guess.astype(np.float32)  # making a copy even if initial_guess is in float32 already
        measurement_max = measurement_vector.max()
        # normalising and converting to float32
        measurement_vector = (measurement_vector / measurement_max).astype(np.float32)
        measurement_squared = np.dot(measurement_vector, measurement_vector)
        y_hat_vector = np.empty_like(measurement_vector)  # host y_hat
        queue = cl.CommandQueue(self._cl_context)

        # copying initial guess and measurement_vector to device
        cl.enqueue_copy(queue, self._solution_device, solution)
        cl.enqueue_copy(queue, self._detectors_device, measurement_vector)

        # calculating y_hat on device
        self._calc_y_hat(queue)

        # starting iterations
        convergence = []
        conv_tol = np.float32(conv_tol)
        success = False
        for k in range(max_iterations):
            # making one iteration on device
            self._make_iteration(queue, relaxation, beta_laplace)

            # calculating y_hat on device
            self._calc_y_hat(queue)

            # copying y_hat to host
            cl.enqueue_copy(queue, y_hat_vector, self._y_hat_device)

            # calculating convergence
            y_hat_squared = np.dot(y_hat_vector, y_hat_vector)
            convergence.append((measurement_squared - y_hat_squared) / measurement_squared)
            time_passed = timer() - time_start

            # checking conditions
            if k > 0 and np.abs(convergence[k] - convergence[k - 1]) < conv_tol:
                success = True
                if self._verbose:
                    print('Convergence limit is reached in {:.4f} s with {} iterations.'.format(time_passed, k + 1))
                break
            if time_passed > time_limit:
                # if no success, the user must be informed even if verbose is False
                print('The time limit has been exceeded, but the convergence limit has not been reached.')
                break

        if (not success) and k == max_iterations - 1:
            # if no success, the user must be informed even if verbose is False
            print('Maximum number of iterations is reached. Time passed: {:.4f} s.'.format(time_passed))

        # copying solution to host
        cl.enqueue_copy(queue, solution, self._solution_device)

        return solution * measurement_max, convergence

    def _calc_y_hat(self, queue):
        if self._use_atomic:
            self._cl_prog.zero_all(queue, self._global_work_size['trivial_detectors'], self._local_work_size['default'],
                                   self._y_hat_device, np.uint32(self._m_detectors))
        if self._geometry_matric_col_maj_device is None:
            self._cl_prog.mat_vec_mult_row_major(queue, self._global_work_size['mult'], self._local_work_size['mult'],
                                                 self._geometry_matrix_device, self._solution_device, self._y_hat_device,
                                                 np.uint32(self._m_detectors), np.uint32(self._n_sources))
        else:
            self._cl_prog.mat_vec_mult_col_major(queue, self._global_work_size['mult'], self._local_work_size['mult'],
                                                 self._geometry_matric_col_maj_device, self._solution_device, self._y_hat_device,
                                                 np.uint32(self._m_detectors), np.uint32(self._n_sources))

    def _make_iteration(self, queue, relaxation, beta_laplace):
        if self._laplacian_matrix_device is not None:
            if self._use_atomic:
                self._cl_prog.zero_all(queue, self._global_work_size['trivial_sources'], self._local_work_size['default'],
                                       self._grad_penalty_device, np.uint32(self._n_sources))
            self._cl_prog.mat_vec_mult_col_major(queue, self._global_work_size['grad'], self._local_work_size['default'],
                                                 self._laplacian_matrix_device, self._solution_device, self._grad_penalty_device,
                                                 np.uint32(self._n_sources), np.uint32(self._n_sources))
            self._cl_prog.vec_scalar_mult(queue, self._global_work_size['trivial_sources'], self._local_work_size['default'],
                                          self._grad_penalty_device, np.float32(beta_laplace), np.uint32(self._n_sources))
        # grad_penalty is just an all-zero array if laplacian_matrix is not provided on initialisation
        self._cl_prog.sart_iteration(queue, self._global_work_size['iter'], self._local_work_size['default'],
                                     self._geometry_matrix_device, self._cell_ray_densities_device, self._ray_lengths_device,
                                     self._y_hat_device, self._detectors_device, self._solution_device, self._grad_penalty_device,
                                     np.float32(relaxation), np.uint32(self._n_sources), np.uint32(self._m_detectors))
        if self._use_atomic:
            self._cl_prog.zero_negative(queue, self._global_work_size['trivial_sources'], self._local_work_size['default'],
                                        self._solution_device, np.uint32(self._n_sources))



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


from raysect.optical cimport Vector3D, Point3D
from raysect.optical.spectrum cimport Spectrum

from cherab.core cimport Plasma
from cherab.core.laser.profile cimport LaserProfile
from cherab.core.laser.laserspectrum cimport LaserSpectrum


cdef class LaserModel:
    cdef:
        Plasma _plasma
        LaserSpectrum _laser_spectrum
        LaserProfile _laser_profile

    cpdef Spectrum emission(self, Point3D point_plasma, Vector3D observation_plasma, Point3D point_laser,
                            Vector3D observation_laser, Spectrum spectrum)

    cdef object __weakref__



# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.optical cimport Vector3D

from cherab.core.math cimport Function3D, VectorFunction3D


cdef class DistributionFunction:

    cdef object notifier

    cdef double evaluate(self, double x, double y, double z, double vx, double vy, double vz) except? -1e999

    cpdef Vector3D bulk_velocity(self, double x, double y, double z)

    cpdef double effective_temperature(self, double x, double y, double z) except? -1e999

    cpdef double density(self, double x, double y, double z) except? -1e999


cdef class ZeroDistribution(DistributionFunction):
    pass


cdef class Maxwellian(DistributionFunction):

    cdef readonly:
        Function3D _density, _temperature
        VectorFunction3D _velocity
        double _atomic_mass




# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from raysect.primitive.mesh.mesh cimport Mesh

from .toroidal_mesh import toroidal_mesh_from_polygon


cpdef Mesh axisymmetric_mesh_from_polygon(object polygon, int num_toroidal_segments=500):
    """
    Generates an Raysect Mesh primitive from the specified 2D polygon.

    :param object polygon: An object which can be converted to a numpy array with shape [N,2] 
                           specifying the wall outline polygon in the R-Z plane. The polygon 
                           should not be closed, i.e. vertex i = 0 and i = N should not be the
                           same vertex, but neighbours.
    :param int num_toroidal_segments: The number of repeating toroidal segments that will be used
                                      to construct the mesh.
    :return: A Raysect Mesh primitive constructed from the R-Z polygon using symmetry.

    .. code-block:: pycon

        >>> from cherab.tools.primitives import axisymmetric_mesh_from_polygon
        >>>
        >>> # wall_polygon is your (N, 2) ndarray describing the polygon
        >>> mesh = axisymmetric_mesh_from_polygon(wall_polygon)
    """

    return toroidal_mesh_from_polygon(polygon, toroidal_extent=360, num_toroidal_segments=num_toroidal_segments)




# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

"""
This demo plots core, edge and blended Generomak 2D plasma profiles.
"""

import numpy as np
from matplotlib.colors import SymLogNorm
from matplotlib.collections import PolyCollection
from matplotlib import pyplot as plt

from cherab.core.math import sample2d
from cherab.core.utility import RecursiveDict

from cherab.generomak.equilibrium import load_equilibrium
from cherab.generomak.plasma.plasma import get_core_interpolators, load_edge_profiles, get_full_profiles


def plot_profiles(core_profile, edge_mesh, edge_data, full_profile, label):

    # get grid parameters
    vertex_coords = np.asarray(edge_mesh["vertex_coords"])
    triangles = np.asarray(edge_mesh["triangles"])
    rl, ru = (vertex_coords[:, 0].min(), vertex_coords[:, 0].max())
    zl, zu = (vertex_coords[:, 1].min(), vertex_coords[:, 1].max())
    nr = 288
    nz = 647

    edge_data = np.asarray(edge_data)

    # Sample core profile on a regular grid
    _, _, core_profile_samples = sample2d(core_profile, (rl, ru, nr), (zl, zu, nz))

    # Sample blended profile on a regular grid
    _, _, profile_samples = sample2d(full_profile, (rl, ru, nr), (zl, zu, nz))

    fig = plt.figure(figsize=(9.5, 5.), tight_layout=True)

    vmax = profile_samples.max()
    linthresh = 0.01 * min(core_profile_samples.max(), edge_data.max())
    color_norm = SymLogNorm(linthresh, vmin=0, vmax=vmax)

    core_profile_samples[core_profile_samples == 0] = np.nan
    profile_samples[profile_samples == 0] = np.nan

    ax_core = fig.add_subplot(131)
    ax_core.imshow(core_profile_samples.T, extent=[rl, ru, zl, zu], origin='lower', norm=color_norm, cmap='gnuplot')
    ax_core.text(0.99, 0.99, 'Core', ha='right', va='top', transform=ax_core.transAxes)
    ax_core.set_xlim(rl, ru)
    ax_core.set_ylim(zl, zu)
    ax_core.set_xlabel('R, m')
    ax_core.set_ylabel('Z, m')

    ax_edge = fig.add_subplot(132, sharex=ax_core, sharey=ax_core)
    collection = PolyCollection(vertex_coords[triangles], norm=color_norm, cmap='gnuplot')
    collection.set_array(edge_data)
    ax_edge.add_collection(collection)
    ax_edge.text(0.99, 0.99, 'Edge', ha='right', va='top', transform=ax_edge.transAxes)
    ax_edge.set_aspect(1)
    ax_edge.set_xlabel('R, m')

    ax_blend = fig.add_subplot(133, sharex=ax_core, sharey=ax_core)
    img = ax_blend.imshow(profile_samples.T, extent=[rl, ru, zl, zu], origin='lower', norm=color_norm, cmap='gnuplot')
    ax_blend.text(0.99, 0.99, 'Blended', ha='right', va='top', transform=ax_blend.transAxes)
    ax_blend.set_xlabel('R, m')

    fig.colorbar(img, label=label)

    return fig


# load Generomak equilibrium
equilibrium = load_equilibrium()

# load 1D core profiles, f(psi_norm)
core_profiles_1d = get_core_interpolators()

# load 2D edge profiles defined on a quadrilateral mesh
edge_data = load_edge_profiles()

# load 2D plasma profiles covering both core and edge regions
# see the source code for the get_full_profiles() to learn how to blend core and edge profiles using a mask function
full_profiles = get_full_profiles(equilibrium, core_profiles_1d)

# map core profiles to 2D using the equilibrium
core_profiles_2d = RecursiveDict()

core_profiles_2d["electron"]["temperature"] = equilibrium.map2d(core_profiles_1d["electron"]["f1d_temperature"])
core_profiles_2d["electron"]["density"] = equilibrium.map2d(core_profiles_1d["electron"]["f1d_density"])

for element, states in core_profiles_1d["composition"].items():
    for charge, state in states.items():
        core_profiles_2d["composition"][element][charge]["density"] = equilibrium.map2d(state["f1d_density"])
        core_profiles_2d["composition"][element][charge]["temperature"] = equilibrium.map2d(state["f1d_temperature"])

core_profiles_2d = core_profiles_2d.freeze()

# Plotting plasma profiles
plot_profiles(core_profiles_2d["electron"]["density"], edge_data["mesh"],
              edge_data["electron"]["density"], full_profiles["electron"]["density"],
              'Electron density, m-3')

plot_profiles(core_profiles_2d["electron"]["temperature"], edge_data["mesh"],
              edge_data["electron"]["temperature"], full_profiles["electron"]["temperature"],
              'Electron temperature, eV')

plot_profiles(core_profiles_2d["composition"]["hydrogen"][1]["temperature"], edge_data["mesh"],
              edge_data["composition"]["hydrogen"][1]["temperature"],
              full_profiles["composition"]["hydrogen"][1]["temperature"], 'Ion temperature, eV')

for element, states in core_profiles_2d["composition"].items():
    for charge, state in states.items():
        state_str = ' {}+'.format(charge) if charge else ' 0'
        plot_profiles(state["density"], edge_data["mesh"],
                      edge_data["composition"][element][charge]["density"],
                      full_profiles["composition"][element][charge]["density"],
                      '{} density, m-3'.format(element + state_str))

plot_profiles(core_profiles_2d["composition"]["hydrogen"][0]["temperature"], edge_data["mesh"],
              edge_data["composition"]["hydrogen"][0]["temperature"],
              full_profiles["composition"]["hydrogen"][0]["temperature"], 'neutral hydrogen effective temperature, eV')

plot_profiles(core_profiles_2d["composition"]["carbon"][0]["temperature"], edge_data["mesh"],
              edge_data["composition"]["carbon"][0]["temperature"],
              full_profiles["composition"]["carbon"][0]["temperature"], 'neutral carbon effective temperature, eV')

plt.show()



from cherab.core.model.laser.laserspectrum import ConstantSpectrum, GaussianSpectrum
from cherab.core.model.laser.profile import UniformEnergyDensity, ConstantAxisymmetricGaussian
from cherab.core.model.laser.profile import ConstantBivariateGaussian, TrivariateGaussian, GaussianBeamAxisymmetric
from cherab.core.model.laser.model import SeldenMatobaThomsonSpectrum


# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from numpy import ndarray
from raysect.optical.observer import TargettedPixel

from .base import Observer0DGroup


class TargettedPixelGroup(Observer0DGroup):
    """
    A group of targetted pixel under a single scene-graph node.

    A scene-graph object regrouping a series of 'TargettedPixel'
    observers as a scene-graph parent. Allows combined observation and display
    control simultaneously.

    :ivar list x_width: Width of pixel along local x axis
    :ivar list y_width: Width of pixel along local y axis
    :ivar list targets: Targets for preferential sampling
    :ivar list targetted_path_prob: Probability of ray being casted at the target
    """
    _OBSERVER_TYPE = TargettedPixel

    @property
    def x_width(self):
        return [pixel.x_width for pixel in self._observers]

    @x_width.setter
    def x_width(self, value):
        if isinstance(value, (list, tuple, ndarray)):
            if len(value) == len(self._observers):
                for pixel, v in zip(self._observers, value):
                    pixel.x_width = v
            else:
                raise ValueError("The length of 'x_width' ({}) "
                                 "mismatches the number of pixels ({}).".format(len(value), len(self._observers)))
        else:
            for pixel in self._observers:
                pixel.x_width = value

    @property
    def y_width(self):
        return [pixel.y_width for pixel in self._observers]

    @y_width.setter
    def y_width(self, value):
        if isinstance(value, (list, tuple, ndarray)):
            if len(value) == len(self._observers):
                for pixel, v in zip(self._observers, value):
                    pixel.y_width = v
            else:
                raise ValueError("The length of 'y_width' ({}) "
                                 "mismatches the number of pixels ({}).".format(len(value), len(self._observers)))
        else:
            for pixel in self._observers:
                pixel.y_width = value

    @property
    def targets(self):
        """
        List of target lists used by pixels for preferential sampling

        :param list value: List of primitives to be set to each pixel or 
                           list of lists containing targets specific for each pixel
                           in this case the number of lists must match number of pixels

        :rtype: list
        """
        return [pixel.targets for pixel in self._observers]

    @targets.setter
    def targets(self, value):
        if all(isinstance(v, (list, tuple)) for v in value):
            if len(value) == len(self._observers):
                for pixel, v in zip(self._observers, value):
                    pixel.targets = v
            else:
                raise ValueError("The number of provided target lists' ({}) "
                                 "mismatches the number of pixels ({}).".format(len(value), len(self._observers)))
        else:
            # assuming a list of primitives, the pixel's setter will throw an error if not
            for pixel in self._observers:
                pixel.targets = value

    @property
    def targetted_path_prob(self):
        return [pixel.targetted_path_prob for pixel in self._observers]
    
    @targetted_path_prob.setter
    def targetted_path_prob(self, value):
        if isinstance(value, (list, tuple)):
            if len(value) == len(self._observers):
                for pixel, v in zip(self._observers, value):
                    pixel.targetted_path_prob = v
            else:
                raise ValueError("The length of 'value' ({}) "
                                 "mismatches the number of pixels ({}).".format(len(value), len(self._observers)))
        else:
            for pixel in self._observers:
                pixel.targetted_path_prob = value



# cython: language_level=3

# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

from cherab.core.math.function cimport autowrap_function2d, autowrap_function3d


cdef class Slice2D(Function1D):
    """
    Exposes a slice of a Function2D as a Function1D.

    :param Function2D function: the function to be sliced.
    :param axis: the axis to be sliced. Must be ['x', 'y'] or [0, 1].
    :param float value: the axis value at which to return a slice

    .. code-block:: pycon

       >>> from cherab.core.math import Slice2D
       >>>
       >>> def f1(x, y):
       >>>     return x**2 + y
       >>>
       >>> f2 = Slice2D(f1, 'x', 1.5)
       >>>
       >>> f2(0)
       2.25
       >>> f2(1)
       3.25
    """

    def __init__(self, object function, object axis, double value):

        # convert string axis to numerical axis value
        if isinstance(axis, str):
            map = {'x': 0, 'y': 1}
            try:
                axis = map[axis.lower()]
            except KeyError:
                raise ValueError('The axis must be either the string \'x\' or \'y\', or the value 0 or 1.')

        # check numerical value
        if axis not in [0, 1]:
            raise ValueError('The axis must be either the string \'x\' or \'y\', or the value 0 or 1.')

        self.axis = axis
        self.value = value
        self._function = autowrap_function2d(function)

    cdef double evaluate(self, double x) except? -1e999:

        if self.axis == 0:
            return self._function.evaluate(self.value, x)
        else:
            return self._function.evaluate(x, self.value)


cdef class Slice3D(Function2D):
    """
    Exposes a slice of a Function3D as a Function2D.

    :param Function3D function: the function to be sliced.
    :param axis: the axis to be sliced. Must be ['x', 'y', 'z'] or [0, 1, 2].
    :param float value: the axis value at which to return a slice

    .. code-block:: pycon

       >>> from cherab.core.math import Slice3D
       >>>
       >>> def f1(x, y, z):
       >>>     return x**3 + y**2 + z
       >>>
       >>> f2 = Slice3D(f1, 'x', 1.5)
       >>>
       >>> f2(0, 0)
       3.375
       >>> f2(1, 0)
       4.375
    """

    def __init__(self, object function, object axis, double value):

        # convert string axis to numerical axis value
        if isinstance(axis, str):
            map = {'x': 0, 'y': 1, 'z': 2}
            try:
                axis = map[axis.lower()]
            except KeyError:
                raise ValueError('The axis must be either the string \'x\', \'y\' or \'z\', or the value 0, 1 or 2.')

        # check numerical value
        if axis not in [0, 1, 2]:
            raise ValueError('The axis must be either the string \'x\', \'y\' or \'z\', or the value 0, 1 or 2.')

        self.axis = axis
        self.value = value
        self._function = autowrap_function3d(function)

    cdef double evaluate(self, double x, double y) except? -1e999:

        if self.axis == 0:
            return self._function.evaluate(self.value, x, y)
        if self.axis == 1:
            return self._function.evaluate(x, self.value, y)
        else:
            return self._function.evaluate(x, y, self.value)



# Copyright 2016-2022 Euratom
# Copyright 2016-2022 United Kingdom Atomic Energy Authority
# Copyright 2016-2022 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

"""
This demo does the same plots as the plot_2d_profiles.py demo, but samples the profiles from the Plasma objects.
"""

import numpy as np
from matplotlib.colors import SymLogNorm
from matplotlib import pyplot as plt

from cherab.core.math import sample3d
from cherab.core.atomic.elements import hydrogen, carbon

from cherab.generomak.plasma import get_core_plasma, get_edge_plasma, get_plasma


def plot_profiles(core_profile, edge_profile, full_profile, r_range, z_range, label):

    # Sample core profile on a regular grid
    _, _, _, core_profile_samples = sample3d(core_profile, r_range, (0, 0, 1), z_range)
    core_profile_samples = core_profile_samples.squeeze()

    # Sample edge profile on a regular grid
    _, _, _, edge_profile_samples = sample3d(edge_profile, r_range, (0, 0, 1), z_range)
    edge_profile_samples = edge_profile_samples.squeeze()

    # Sample blended profile on a regular grid
    _, _, _, full_profile_samples = sample3d(full_profile, r_range, (0, 0, 1), z_range)
    full_profile_samples = full_profile_samples.squeeze()

    fig = plt.figure(figsize=(9.5, 5.), tight_layout=True)

    vmax = full_profile_samples.max()
    linthresh = 0.01 * min(core_profile_samples.max(), edge_profile_samples.max())
    color_norm = SymLogNorm(linthresh, vmin=0, vmax=vmax)

    core_profile_samples[core_profile_samples == 0] = np.nan
    edge_profile_samples[edge_profile_samples == 0] = np.nan
    full_profile_samples[full_profile_samples == 0] = np.nan

    ax_core = fig.add_subplot(131)
    ax_core.imshow(core_profile_samples.T, extent=[r_range[0], r_range[1], z_range[0], z_range[1]], origin='lower', norm=color_norm, cmap='gnuplot')
    ax_core.text(0.99, 0.99, 'Core', ha='right', va='top', transform=ax_core.transAxes)
    ax_core.set_xlim(r_range[0], r_range[1])
    ax_core.set_ylim(z_range[0], z_range[1])
    ax_core.set_xlabel('R, m')
    ax_core.set_ylabel('Z, m')

    ax_edge = fig.add_subplot(132, sharex=ax_core, sharey=ax_core)
    img = ax_edge.imshow(edge_profile_samples.T, extent=[r_range[0], r_range[1], z_range[0], z_range[1]], origin='lower', norm=color_norm, cmap='gnuplot')
    ax_edge.text(0.99, 0.99, 'Edge', ha='right', va='top', transform=ax_edge.transAxes)
    ax_edge.set_xlabel('R, m')

    ax_blend = fig.add_subplot(133, sharex=ax_core, sharey=ax_core)
    img = ax_blend.imshow(full_profile_samples.T, extent=[r_range[0], r_range[1], z_range[0], z_range[1]], origin='lower', norm=color_norm, cmap='gnuplot')
    ax_blend.text(0.99, 0.99, 'Blended', ha='right', va='top', transform=ax_blend.transAxes)
    ax_blend.set_xlabel('R, m')

    fig.colorbar(img, label=label)

    return fig


# get Generomak core plasma
core_plasma = get_core_plasma()

# get Generomak edge plasma
edge_plasma = get_edge_plasma()

# get Generomak full plasma (blended core and edge profiles)
full_plasma = get_plasma()

# plasma domain
r_range = (0.78, 2.23, 290)
z_range = (-1.74, 1.49, 647)

# Plotting plasma profiles
plot_profiles(core_plasma.electron_distribution.density,
              edge_plasma.electron_distribution.density,
              full_plasma.electron_distribution.density,
              r_range, z_range, 'Electron density, m-3')

plot_profiles(core_plasma.electron_distribution.effective_temperature,
              edge_plasma.electron_distribution.effective_temperature,
              full_plasma.electron_distribution.effective_temperature,
              r_range, z_range, 'Electron temperature, eV')

plot_profiles(core_plasma.composition.get(hydrogen, 1).distribution.effective_temperature,
              edge_plasma.composition.get(hydrogen, 1).distribution.effective_temperature,
              full_plasma.composition.get(hydrogen, 1).distribution.effective_temperature,
              r_range, z_range, 'Ion temperature, eV')

for element, charges in ((hydrogen, (0, 1)), (carbon, (0, 1, 2, 3, 4, 5, 6))):
    for charge in charges:
        state_str = ' {}+'.format(charge) if charge else ' 0'
        plot_profiles(core_plasma.composition.get(element, charge).distribution.density,
                      edge_plasma.composition.get(element, charge).distribution.density,
                      full_plasma.composition.get(element, charge).distribution.density,
                      r_range, z_range, '{} density, m-3'.format(element.name + state_str))

plot_profiles(core_plasma.composition.get(hydrogen, 0).distribution.effective_temperature,
              edge_plasma.composition.get(hydrogen, 0).distribution.effective_temperature,
              full_plasma.composition.get(hydrogen, 0).distribution.effective_temperature,
              r_range, z_range, 'neutral hydrogen effective temperature, eV')

plot_profiles(core_plasma.composition.get(carbon, 0).distribution.effective_temperature,
              edge_plasma.composition.get(carbon, 0).distribution.effective_temperature,
              full_plasma.composition.get(carbon, 0).distribution.effective_temperature,
              r_range, z_range, 'neutral carbon effective temperature, eV')

plt.show()



# cython: language_level=3

# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence

from numpy cimport ndarray
from raysect.optical cimport World, Primitive, Ray, Spectrum, Point3D, Vector3D, AffineMatrix3D
from raysect.optical.material cimport VolumeIntegrator, InhomogeneousVolumeEmitter


cdef class RayTransferIntegrator(VolumeIntegrator):

    cdef:
        double _step
        int _min_samples


cdef class CylindricalRayTransferIntegrator(RayTransferIntegrator):

    cpdef Spectrum integrate(self, Spectrum spectrum, World world, Ray ray, Primitive primitive,
                             InhomogeneousVolumeEmitter material, Point3D start_point, Point3D end_point,
                             AffineMatrix3D world_to_primitive, AffineMatrix3D primitive_to_world)


cdef class CartesianRayTransferIntegrator(RayTransferIntegrator):

    cpdef Spectrum integrate(self, Spectrum spectrum, World world, Ray ray, Primitive primitive,
                             InhomogeneousVolumeEmitter material, Point3D start_point, Point3D end_point,
                             AffineMatrix3D world_to_primitive, AffineMatrix3D primitive_to_world)


cdef class RayTransferEmitter(InhomogeneousVolumeEmitter):

    cdef:
        int[3] _grid_shape
        double[3] _grid_steps
        int _bins
        ndarray _voxel_map
        public:
            int[:, :, ::1] voxel_map_mv

    cdef ndarray _map_from_mask(self, mask)


cdef class CylindricalRayTransferEmitter(RayTransferEmitter):

    cdef:
        double _dr, _dphi, _dz, _period, _rmin

    cpdef Spectrum emission_function(self, Point3D point, Vector3D direction, Spectrum spectrum,
                                     World world, Ray ray, Primitive primitive,
                                     AffineMatrix3D world_to_primitive, AffineMatrix3D primitive_to_world)


cdef class CartesianRayTransferEmitter(RayTransferEmitter):

    cdef:
        double _dx, _dy, _dz

    cpdef Spectrum emission_function(self, Point3D point, Vector3D direction, Spectrum spectrum,
                                     World world, Ray ray, Primitive primitive,
                                     AffineMatrix3D world_to_primitive, AffineMatrix3D primitive_to_world)


# Copyright 2016-2024 Euratom
# Copyright 2016-2024 United Kingdom Atomic Energy Authority
# Copyright 2016-2024 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np

from cherab.core.utility.conversion import PhotonToJ

cimport cython
from libc.math cimport INFINITY, log10
from raysect.core.math.function.float cimport Interpolator1DArray, Interpolator2DArray, Constant2D, Arg2D
from cherab.core.math cimport IsoMapper2D

# todo: clarify variables

cdef class BeamStoppingRate(CoreBeamStoppingRate):
    """
    The beam stopping coefficient interpolation class.

    Data is interpolated with cubic spline in log-log space.
    Linear and quadratic extrapolations are used for "sen" and "st" respectively
    if extrapolate is True.

    :param dict data: A beam stopping rate dictionary containing the following entries:

    |      'e': 1D array of size (N) with interaction energy in eV/amu,
    |      'n': 1D array of size (M) with target electron density in m^-3,
    |      't': 1D array of size (K) with target electron temperature in eV,
    |      'sen': 2D array of size (N, M) with beam stopping rate energy component in m^3.s^-1.
    |      'st': 1D array of size (K) with beam stopping rate temperature component in m^3.s^-1.
    |      'sref': reference beam stopping rate in m^3.s^-1.
    |  The total beam stopping rate: s = sen * st / sref.

    :param bint extrapolate: Set to True to enable extrapolation, False to disable (default).

    :ivar tuple beam_energy_range: Interaction energy interpolation range.
    :ivar tuple density_range: Target electron density interpolation range.
    :ivar tuple temperature_range: Target electron temperature interpolation range.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    @cython.cdivision(True)
    def __init__(self, dict data, bint extrapolate=False):

        self.raw_data = data

        # unpack
        e = data["e"]                          # eV/amu
        n = data["n"]                          # m^-3
        t = data["t"]                          # eV
        sen = np.log10(data["sen"])                      # m^3/s
        st = np.log10(data["st"] / data["sref"])         # dimensionless

        # store limits of data
        self.beam_energy_range = e.min(), e.max()
        self.density_range = n.min(), n.max()
        self.temperature_range = t.min(), t.max()

        # interpolate
        extrapolation_type_2d = 'linear' if extrapolate else 'none'
        extrapolation_type_1d = 'quadratic' if extrapolate else 'none'
        if len(e) == 1 and len(n) == 1:
            self._npl_eb = Constant2D(sen[0, 0])
        elif len(e) == 1:
            self._npl_eb = IsoMapper2D(Arg2D('y'), Interpolator1DArray(np.log10(n), sen[0], 'cubic', extrapolation_type_1d, INFINITY))
        elif len(n) == 1:
            self._npl_eb = IsoMapper2D(Arg2D('x'), Interpolator1DArray(np.log10(e), sen[:, 0], 'cubic', extrapolation_type_1d, INFINITY))
        else:
            self._npl_eb = Interpolator2DArray(np.log10(e), np.log10(n), sen, 'cubic', extrapolation_type_2d, INFINITY, INFINITY)
        self._tp = Interpolator1DArray(np.log10(t), st, 'cubic', extrapolation_type_1d, INFINITY)

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        """
        Interpolates and returns the beam coefficient for the supplied parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param density: Target electron density in m^-3
        :param temperature: Target temperature in eV.
        :return: The beam stopping coefficient in m^3.s^-1
        """

        # need to handle zeros, also density and temperature can become negative due to cubic interpolation
        if energy <= 0 or density <= 0 or temperature <= 0:
            return 0

        # calculate rate and convert from log10 space to linear space
        return 10 ** (self._npl_eb.evaluate(log10(energy), log10(density)) + self._tp.evaluate(log10(temperature)))


cdef class NullBeamStoppingRate(CoreBeamStoppingRate):
    """
    A beam stopping rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        return 0.0


cdef class BeamPopulationRate(CoreBeamPopulationRate):
    """
    The beam population coefficient interpolation class.

    Data is interpolated with cubic spline in log-log space.
    Linear and quadratic extrapolations are used for "sen" and "st" respectively
    if extrapolate is True.

    :param dict data: Beam population rate dictionary containing the following entries:

    |      'e': 1D array of size (N) with interaction energy in eV/amu,
    |      'n': 1D array of size (M) with target electron density in m^-3,
    |      't': 1D array of size (K) with target electron temperature in eV,
    |      'sen': 2D array of size (N, M) with dimensionless beam population rate energy component.
    |      'st': 1D array of size (K) with dimensionless beam population rate temperature component.
    |      'sref': reference dimensionless beam population rate.
    |  The total beam population rate: s = sen * st / sref.

    :param bint extrapolate: Set to True to enable extrapolation, False to disable (default).

    :ivar tuple beam_energy_range: Interaction energy interpolation range.
    :ivar tuple density_range: Target electron density interpolation range.
    :ivar tuple temperature_range: Target electron temperature interpolation range.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    @cython.cdivision(True)
    def __init__(self, dict data, bint extrapolate=False):

        self.raw_data = data

        # unpack
        e = data["e"]                          # eV/amu
        n = data["n"]                          # m^-3
        t = data["t"]                          # eV
        sen = np.log10(data["sen"])                      # dimensionless
        st = np.log10(data["st"] / data["sref"])         # dimensionless

        # store limits of data
        self.beam_energy_range = e.min(), e.max()
        self.density_range = n.min(), n.max()
        self.temperature_range = t.min(), t.max()

        # interpolate
        extrapolation_type_2d = 'linear' if extrapolate else 'none'
        extrapolation_type_1d = 'quadratic' if extrapolate else 'none'
        if len(e) == 1 and len(n) == 1:
            self._npl_eb = Constant2D(sen[0, 0])
        elif len(e) == 1:
            self._npl_eb = IsoMapper2D(Arg2D('y'), Interpolator1DArray(np.log10(n), sen[0], 'cubic', extrapolation_type_1d, INFINITY))
        elif len(n) == 1:
            self._npl_eb = IsoMapper2D(Arg2D('x'), Interpolator1DArray(np.log10(e), sen[:, 0], 'cubic', extrapolation_type_1d, INFINITY))
        else:
            self._npl_eb = Interpolator2DArray(np.log10(e), np.log10(n), sen, 'cubic', extrapolation_type_2d, INFINITY, INFINITY)
        self._tp = Interpolator1DArray(np.log10(t), st, 'cubic', extrapolation_type_1d, INFINITY)

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        """
        Interpolates and returns the beam coefficient for the supplied parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param density: Target electron density in m^-3
        :param temperature: Target temperature in eV.
        :return: The beam population coefficient in dimensionless units.
        """

        # need to handle zeros, also density and temperature can become negative due to cubic interpolation
        if energy <= 0 or density <= 0 or temperature <= 0:
            return 0

        # calculate rate and convert from log10 space to linear space
        return 10 ** (self._npl_eb.evaluate(log10(energy), log10(density)) + self._tp.evaluate(log10(temperature)))


cdef class NullBeamPopulationRate(CoreBeamPopulationRate):
    """
    A beam population rate that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        return 0.0


cdef class BeamEmissionPEC(CoreBeamEmissionPEC):
    """
    The beam emission coefficient interpolation class.

    Data is interpolated with cubic spline in log-log space.
    Linear and quadratic extrapolations are used for "sen" and "st" respectively
    if extrapolate is True.

    :param dict data: Beam emission rate dictionary containing the following entries:

    |      'e': 1D array of size (N) with interaction energy in eV/amu,
    |      'n' 1D array of size (M) with target electron density in m^-3,
    |      't' 1D array of size (K) with target electron temperature in eV,
    |      'sen' 2D array of size (N, M) with beam emission rate energy component in photon.m^3.s^-1.
    |      'st' 1D array of size (K) with beam emission rate temperature component in photon.m^3.s^-1.
    |      'sref': reference beam emission rate in photon.m^3.s^-1.

    :param double wavelength: The natural wavelength of the emission line associated with the rate data in nm.
    :param bint extrapolate: Set to True to enable extrapolation, False to disable (default).

    :ivar tuple beam_energy_range: Interaction energy interpolation range.
    :ivar tuple density_range: Target electron density interpolation range.
    :ivar tuple temperature_range: Target electron temperature interpolation range.
    :ivar dict raw_data: Dictionary containing the raw data.
    """

    @cython.cdivision(True)
    def __init__(self, dict data, double wavelength, bint extrapolate=False):

        self.wavelength = wavelength
        self.raw_data = data

        # unpack
        e = data["e"]                                   # eV/amu
        n = data["n"]                                   # m^-3
        t = data["t"]                                   # eV
        sen = np.log10(PhotonToJ.to(data["sen"], wavelength))     # W.m^3
        st = np.log10(data["st"] / data["sref"])                  # dimensionless

        # store limits of data
        self.beam_energy_range = e.min(), e.max()
        self.density_range = n.min(), n.max()
        self.temperature_range = t.min(), t.max()

        # interpolate
        extrapolation_type_2d = 'linear' if extrapolate else 'none'
        extrapolation_type_1d = 'quadratic' if extrapolate else 'none'
        if len(e) == 1 and len(n) == 1:
            self._npl_eb = Constant2D(sen[0, 0])
        elif len(e) == 1:
            self._npl_eb = IsoMapper2D(Arg2D('y'), Interpolator1DArray(np.log10(n), sen[0], 'cubic', extrapolation_type_1d, INFINITY))
        elif len(n) == 1:
            self._npl_eb = IsoMapper2D(Arg2D('x'), Interpolator1DArray(np.log10(e), sen[:, 0], 'cubic', extrapolation_type_1d, INFINITY))
        else:
            self._npl_eb = Interpolator2DArray(np.log10(e), np.log10(n), sen, 'cubic', extrapolation_type_2d, INFINITY, INFINITY)
        self._tp = Interpolator1DArray(np.log10(t), st, 'cubic', extrapolation_type_1d, INFINITY)

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        """
        Interpolates and returns the beam coefficient for the supplied parameters.

        If the requested data is out-of-range then the call with throw a ValueError exception.

        :param energy: Interaction energy in eV/amu.
        :param density: Target electron density in m^-3
        :param temperature: Target temperature in eV.
        :return: The beam emission coefficient in m^3.s^-1
        """

        # need to handle zeros, also density and temperature can become negative due to cubic interpolation
        if energy <= 0 or density <= 0 or temperature <= 0:
            return 0

        # calculate rate and convert from log10 space to linear space
        return 10 ** (self._npl_eb.evaluate(log10(energy), log10(density)) + self._tp.evaluate(log10(temperature)))


cdef class NullBeamEmissionPEC(CoreBeamEmissionPEC):
    """
    A beam emission PEC that always returns zero.
    Needed for use cases where the required atomic data is missing.
    """

    cpdef double evaluate(self, double energy, double density, double temperature) except? -1e999:
        return 0.0




# Copyright 2016-2018 Euratom
# Copyright 2016-2018 United Kingdom Atomic Energy Authority
# Copyright 2016-2018 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import numpy as np
from scipy import linalg


def invert_svd(w_matrix, b_vector):
    """
    Performs a Singular Value Decomposition (SVD) operation inversion.

    :param np.ndarray w_matrix: The sensitivity matrix describing the coupling between the
      detectors and the voxels. Must be an array with shape :math:`(N_d, N_s)`.
    :param np.ndarray b_vector: The measured power/radiance vector with shape :math:`(N_d)`.
    :return: The solution vector x as an ndarray.
    """

    # Compute the Moore-Penrose pseudo-inverse of a matrix from SVD
    inverse_w_matrix = np.matrix(linalg.pinv(w_matrix))

    # reshape b_vector into a column vector
    b_vector = b_vector.reshape((len(b_vector), 1))

    inverted_x_vector = (inverse_w_matrix * b_vector).flatten()

    return np.asarray(inverted_x_vector).flatten()



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.


from raysect.core cimport Primitive
from raysect.optical cimport World, Primitive, Ray, Spectrum, Point3D, Vector3D, AffineMatrix3D
from raysect.optical.material.emitter cimport InhomogeneousVolumeEmitter
from raysect.optical.material.emitter.inhomogeneous cimport VolumeIntegrator

from cherab.core.laser.node cimport Laser
from cherab.core.laser.model cimport LaserModel


cdef class LaserMaterial(InhomogeneousVolumeEmitter):

    def __init__(self, Laser laser not None, Primitive laser_segment not None, list models, VolumeIntegrator integrator not None):

        super().__init__(integrator)

        self._laser = laser
        self._primitive = laser_segment
        self.importance = laser.importance
        
        #validate and set models
        for model in models:
            if not isinstance(model, LaserModel):
                raise TypeError("Model supplied to laser are not LaserMaterial is not LaserModel")
            model.plasma = laser.plasma
            model.laser_profile = laser.laser_profile
            model.laser_spectrum = laser.laser_spectrum

        self._models = models

    cpdef Spectrum emission_function(self, Point3D point, Vector3D direction, Spectrum spectrum,
                                     World world, Ray ray, Primitive primitive,
                                     AffineMatrix3D to_local, AffineMatrix3D to_world):

        cdef:
            Point3D point_plasma, point_laser
            Vector3D direction_plasma, direction_laser
            LaserModel model
        
        # cache the important transforms
        if self._laser_segment_to_laser_node is None or self._laser_to_plasma is None:
            self._cache_transforms()

        point_laser = point.transform(self._laser_segment_to_laser_node)
        direction_laser = direction.transform(self._laser_segment_to_laser_node) # observation vector in the laser frame
        point_plasma = point.transform(self._laser_to_plasma)
        direction_plasma = direction.transform(self._laser_to_plasma)

        for model in self._models:
            spectrum = model.emission(point_plasma, direction_plasma, point_laser, direction_laser, spectrum)

        return spectrum

    cdef void _cache_transforms(self):
        """
        cache transforms from laser primitive to laser and plasma
        """

        # if transforms are cached, the material should be used only for one primitive for safety
        if not len(self.primitives) == 1:
            raise ValueError("LaserMaterial must be attached to exactly one primitive.")

        self._laser_segment_to_laser_node = self._primitive.to(self._laser)
        self._laser_to_plasma = self._primitive.to(self._laser.get_plasma())



# Copyright 2016-2021 Euratom
# Copyright 2016-2021 United Kingdom Atomic Energy Authority
# Copyright 2016-2021 Centro de Investigaciones Energéticas, Medioambientales y Tecnológicas
#
# Licensed under the EUPL, Version 1.1 or – as soon they will be approved by the
# European Commission - subsequent versions of the EUPL (the "Licence");
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at:
#
# https://joinup.ec.europa.eu/software/page/eupl5
#
# Unless required by applicable law or agreed to in writing, software distributed
# under the Licence is distributed on an "AS IS" basis, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations
# under the Licence.

import warnings

import matplotlib.pyplot as plt
from raysect.optical import Spectrum
from raysect.optical.observer import RadiancePipeline0D, PowerPipeline0D, SpectralRadiancePipeline0D, SpectralPowerPipeline0D


def select_pipelines(group, item):
    """
    Selects pipelines of the same type based on index or name from the provided group.

    If name is used, error is raised when more than one pipeline is found in an observer.
    An ValueError is also raised if found pipelines are not of the same type.

    :param Observer0DGroup group: Observer group from which to select pipelines.
    :param str/int item: The index or name of the pipeline to be selected.
    :return list pipelines: matching pipelines
    :return list observers: observers with matching pipelines
    """
    pipelines = []
    observers = []
    for observer in group.observers:
        if isinstance(item, int):
            try:
                pipelines.append(observer.pipelines[item])
                observers.append(observer)
            except IndexError:
                warnings.warn('Invalid pipeline index {} for observer {} with {} pipelines'.format(
                                item, observer, len(observer.pipelines)))
        elif isinstance(item, str):
            matching_pipelines = [pipeline for pipeline in observer.pipelines if pipeline.name == item]
            pipelines_num = len(matching_pipelines)
            if pipelines_num == 0:
                warnings.warn('Found no matching pipelines for name {} in observer {}'.format(item, observer))
            elif pipelines_num == 1:
                pipelines.append(matching_pipelines[0])
                observers.append(observer)
            else:
                raise ValueError("Found {} matching pipelines for name {} in observer {}.".format(
                                pipelines_num, item, observer))

    pipeline_types = set(type(pipeline) for pipeline in pipelines)
    if len(pipeline_types) > 1:
        raise ValueError("Pipelines {} have different types for different observers.".format(item))

    return pipelines, observers


def plot_group_total(group, item=0, ax=None):
    """
    Plots total (wavelength-integrated) signal for each observer in the group.
    
    :param Observer0DGroup group: Group class with observers hodling data to be plotted.
    :param str/int item: The index or name of the pipeline. Default: 0.
    :param Axes ax: Existing matplotlib axes.
    :return Axes ax: Matplotlib axes with plotted spectra
    """
    pipelines, observers = select_pipelines(group, item)

    if ax is None:
        _, ax = plt.subplots(constrained_layout=True)

    signal = []
    tick_labels = []
    for pipeline, observer in zip(pipelines, observers):
        if isinstance(pipeline, SpectralPowerPipeline0D):
            spectrum = Spectrum(pipeline.min_wavelength, pipeline.max_wavelength, pipeline.bins)
            spectrum.samples[:] = pipeline.samples.mean
            signal.append(spectrum.total())
        else:
            signal.append(pipeline.value.mean)

        if observer.name and len(observer.name):
            tick_labels.append(observer.name)
        else:
            tick_labels.append(group.observers.index(observer))

    if isinstance(pipeline, (SpectralRadiancePipeline0D, RadiancePipeline0D)):
        ylabel = 'Radiance [W/m^2/str]'
    else:  # SpectralPowerPipeline0D or PowerPipeline0D
        ylabel = 'Power [W]'

    ax.bar(list(range(len(signal))), signal, tick_label=tick_labels, label=item)

    if isinstance(item, int):
        # check if pipelines share the same name
        if len(set(pipeline.name for pipeline in pipelines)) == 1 and pipelines[0].name and len(pipelines[0].name):
            ax.set_title('{}: {}'.format(group.name, pipelines[0].name))
        else:
            # pipelines have different names or name is not set
            ax.set_title('{}: pipeline {}'.format(group.name, item))
    elif isinstance(item, str):
        ax.set_title('{}: {}'.format(group.name, item))

    ax.set_ylabel(ylabel)
    ax.set_xlabel('Observer')

    return ax


def plot_group_spectra(group, item=0, in_photons=False, ax=None):
    """
    Plot the spectra observed by each observer in the group for a given pipeline.
    
    :param Observer0DGroup group: Group class with observers hodling data to be plotted.
    :param str/int item: The index or name of the pipeline. Default: 0.
    :param bool in_photons: If True, plots the spectrum in photon/s/nm instead of W/nm.
                            Default is False.
    :param Axes ax: Existing matplotlib axes.
    :return Axes ax: Matplotlib axes with plotted spectra
    """
    pipelines, observers = select_pipelines(group, item)

    if ax is None:
        _, ax = plt.subplots(constrained_layout=True)
    
    for pipeline, observer in zip(pipelines, observers):
        spectrum_observed = Spectrum(pipeline.min_wavelength, pipeline.max_wavelength, pipeline.bins)
        spectrum_observed.samples[:] = pipeline.samples.mean
        if in_photons:
            # turn the samples into photon/s
            spectrum = spectrum_observed.new_spectrum()
            spectrum.samples[:] = spectrum_observed.to_photons()
        else:
            spectrum = spectrum_observed

        label = observer.name if observer.name and len(observer.name) else group.observers.index(observer)

        if spectrum.samples.size > 1:
            ax.plot(spectrum.wavelengths, spectrum.samples, label=label)
        else:
            ax.plot(spectrum.wavelengths, spectrum.samples, marker='o', ls='none', label=label)
    if isinstance(pipelines[0], SpectralRadiancePipeline0D):
        ylabel = 'Spectral radiance [photon/s/m^2/str/nm]' if in_photons else 'Spectral radiance [W/m^2/str/nm]'
    else:  # SpectralPowerPipeline0D
        ylabel = 'Spectral power [photon/s/nm]' if in_photons else 'Spectral power [W/nm]'

    if isinstance(item, int):
        # check if pipelines share the same name
        if len(set(pipeline.name for pipeline in pipelines)) == 1 and pipelines[0].name and len(pipelines[0].name):
            ax.set_title('{}: {}'.format(group.name, pipelines[0].name))
        else:
            # pipelines have different names or name is not set
            ax.set_title('{}: pipeline {}'.format(group.name, item))
    elif isinstance(item, str):
        ax.set_title('{}: {}'.format(group.name, item))

    ax.set_xlabel('Wavelength [nm]')
    ax.set_ylabel(ylabel)
    ax.legend()

    return ax